"""
============
rororo.utils
============

Different utility functions, which are common used in web development, like
converting string to int or bool.

"""

from distutils.util import strtobool
from typing import Any, Collection, Optional, Union

from .annotations import T


def ensure_collection(value: Collection[str]) -> Collection[str]:
    """Ensure that given value is a collection, not a single string.

    As passing single string validates ``Collection[str]`` type annotation,
    this function converts given single string to a tuple with one item.

    In all other cases return given value.
    """
    if isinstance(value, str):
        return (value,)
    return value


def to_bool(value: Any) -> bool:
    """Convert string or other Python object to boolean.

    **Rationalle**

    Passing flags is one of the most common cases of using environment vars and
    as values are strings we need to have an easy way to convert them to
    boolean Python value.

    Without this function int or float string values can be converted as false
    positives, e.g. ``bool('0') => True``, but using this function ensure that
    digit flag be properly converted to boolean value.

    :param value: String or other value.
    """
    return bool(strtobool(value) if isinstance(value, str) else value)


def to_int(value: str, default: T = None) -> Union[int, Optional[T]]:
    """Convert given value to int.

    If conversion failed, return default value without raising Exception.

    :param value: Value to convert to int.
    :param default: Default value to use in case of failed conversion.
    """
    try:
        return int(value)
    except (TypeError, ValueError):
        return default



from aiohttp import web

from .constants import REQUEST_CORE_REQUEST_KEY, REQUEST_OPENAPI_CONTEXT_KEY
from .core_data import to_core_openapi_request, to_core_openapi_response
from .core_validators import validate_core_request, validate_core_response
from .data import OpenAPIContext
from .utils import get_openapi_spec


async def validate_request(request: web.Request) -> web.Request:
    config_dict = request.config_dict

    core_request = await to_core_openapi_request(request)
    request[REQUEST_CORE_REQUEST_KEY] = core_request

    security, parameters, data = validate_core_request(
        get_openapi_spec(config_dict), core_request
    )
    request[REQUEST_OPENAPI_CONTEXT_KEY] = OpenAPIContext(
        request=request,
        app=request.app,
        config_dict=config_dict,
        parameters=parameters,
        security=security,
        data=data,
    )

    return request


def validate_response(
    request: web.Request, response: web.StreamResponse
) -> web.StreamResponse:
    validate_core_response(
        get_openapi_spec(request.config_dict),
        request[REQUEST_CORE_REQUEST_KEY],
        to_core_openapi_response(response),
    )
    return response



#: Key to store OpenAPI schema within the ``web.Application`` instance
APP_OPENAPI_SCHEMA_KEY = "rororo_openapi_schema"

#: Key to store OpenAPI spec within the ``web.Application`` instance
APP_OPENAPI_SPEC_KEY = "rororo_openapi_spec"

#: Key to store request method -> operation ID mapping in handler
HANDLER_OPENAPI_MAPPING_KEY = "__rororo_openapi_mapping__"

#: Key to store current OpenAPI core operation within the ``web.Request``
#: instance
REQUEST_CORE_OPERATION_KEY = "rororo_core_operation"

#: Key to store current OpenAPI core request within the ``web.Request``
#: instance
REQUEST_CORE_REQUEST_KEY = "rororo_core_request"

#: Key to store valid OpenAPI context within the ``web.Request`` instance
REQUEST_OPENAPI_CONTEXT_KEY = "rororo_openapi_context"



from typing import Any, cast, Optional, Union

from aiohttp import web
from aiohttp.helpers import ChainMapProxy
from openapi_core.schema.specs.models import Spec
from openapi_core.validation.request.datatypes import OpenAPIRequest
from yarl import URL

from .constants import (
    APP_OPENAPI_SCHEMA_KEY,
    APP_OPENAPI_SPEC_KEY,
    REQUEST_OPENAPI_CONTEXT_KEY,
)
from .data import OpenAPIContext, OpenAPIParameters
from .exceptions import ConfigurationError, ContextError
from ..annotations import DictStrAny


def add_prefix(path: str, prefix: Optional[str]) -> str:
    if prefix:
        if prefix[-1] == "/":
            prefix = prefix[:-1]
        return f"{prefix}{path}"
    return path


def get_base_url(core_request: OpenAPIRequest) -> str:
    return str(URL(core_request.full_url_pattern).with_path("/"))


def get_openapi_context(request: web.Request) -> OpenAPIContext:
    """Shortcut to retrieve OpenAPI schema from ``aiohttp.web`` request.

    ``OpenAPIContext`` attached to :class:`aiohttp.web.Request` instance only
    if current request contains valid data.

    ``ContextError`` raises if, for some reason, the function called outside of
    valid OpenAPI request context.
    """
    try:
        return cast(OpenAPIContext, request[REQUEST_OPENAPI_CONTEXT_KEY])
    except KeyError:
        raise ContextError(
            "Request instance does not contain valid OpenAPI context. In "
            "most cases it means, the function is called outside of valid "
            "OpenAPI request context."
        )


def get_openapi_schema(
    mixed: Union[web.Application, ChainMapProxy]
) -> DictStrAny:
    """Shortcut to retrieve OpenAPI schema from ``aiohttp.web`` application.

    ``ConfigruationError`` raises if :class:`aiohttp.web.Application` does not
    contain registered OpenAPI schema.
    """
    try:
        return cast(DictStrAny, mixed[APP_OPENAPI_SCHEMA_KEY])
    except KeyError:
        raise ConfigurationError(
            "Seems like OpenAPI schema not registered to the application. Use "
            '"from rororo import setup_openapi" function to register OpenAPI '
            "schema to your web.Application."
        )


def get_openapi_spec(mixed: Union[web.Application, ChainMapProxy]) -> Spec:
    """Shortcut to retrieve OpenAPI spec from ``aiohttp.web`` application.

    ``ConfigruationError`` raises if :class:`aiohttp.web.Application` does not
    contain registered OpenAPI spec.
    """
    try:
        return mixed[APP_OPENAPI_SPEC_KEY]
    except KeyError:
        raise ConfigurationError(
            "Seems like OpenAPI spec not registered to the application. Use "
            '"from rororo import setup_openapi" function to register OpenAPI '
            "schema to your web.Application."
        )


def get_validated_data(request: web.Request) -> Any:
    """Shortcut to get validated data (request body) for given request.

    In case when current request has no valid OpenAPI context attached -
    ``ContextError`` will be raised.
    """
    return get_openapi_context(request).data


def get_validated_parameters(request: web.Request) -> OpenAPIParameters:
    """Shortcut to get validated parameters for given request.

    In case when current request has no valid OpenAPI context attached -
    ``ContextError`` will be raised.
    """
    return get_openapi_context(request).parameters



from typing import cast, Optional, Union

from aiohttp import BasicAuth, hdrs
from openapi_core.schema.operations.models import Operation
from openapi_core.schema.security_schemes.enums import (
    HttpAuthScheme,
    SecuritySchemeType,
)
from openapi_core.schema.security_schemes.models import SecurityScheme
from openapi_core.security.exceptions import SecurityError as CoreSecurityError
from openapi_core.validation.request.datatypes import OpenAPIRequest
from openapi_core.validation.request.validators import RequestValidator
from pyrsistent import pmap

from .exceptions import BasicSecurityError, SecurityError
from ..annotations import MappingStrAny


AUTHORIZATION_HEADER = hdrs.AUTHORIZATION


def basic_auth_factory(value: str) -> BasicAuth:
    # Workaround for ``openapi-core==0.13.3``
    if ":" in value:
        return BasicAuth(*(item.strip() for item in value.split(":", 1)))
    # In ``openapi-core==0.13.4`` parsing security schemas changed, now the
    # value for basic auth is base64 encoded string
    return BasicAuth.decode(f"Basic {value}")


def get_jwt_security_data(request: OpenAPIRequest) -> Optional[str]:
    """Get JWT bearer security data.

    At a moment, openapi-core attempts to decode JWT header using same rules
    as for basic auth, which might return unexpected results.
    """
    header: Optional[str] = request.parameters.header.get(AUTHORIZATION_HEADER)

    # Header does not exist
    if header is None:
        return None

    try:
        maybe_bearer, value = header.split(" ", 1)
    except ValueError:
        return None

    if maybe_bearer.lower() != "bearer":
        return None

    return value


def get_security_data(
    validator: RequestValidator, request: OpenAPIRequest, scheme_name: str
) -> Optional[Union[BasicAuth, str]]:
    """Get security data from request.

    Currently supported getting API Key & HTTP security data. OAuth & OpenID
    data not supported yet.
    """
    if is_jwt_bearer_security_scheme(validator, scheme_name):
        return get_jwt_security_data(request)

    try:
        value: str = validator._get_security_value(scheme_name, request)
    except CoreSecurityError:
        return None

    if is_basic_auth_security_scheme(validator, scheme_name):
        return basic_auth_factory(value)
    return value


def get_security_scheme(
    validator: RequestValidator, scheme_name: str
) -> Optional[SecurityScheme]:
    return validator.spec.components.security_schemes.get(scheme_name)


def is_basic_auth_security_scheme(
    validator: RequestValidator, scheme_name: str
) -> bool:
    scheme = get_security_scheme(validator, scheme_name)
    if scheme is None:
        return False
    return cast(
        bool,
        scheme.type == SecuritySchemeType.HTTP
        and scheme.scheme == HttpAuthScheme.BASIC,
    )


def is_jwt_bearer_security_scheme(
    validator: RequestValidator, scheme_name: str
) -> bool:
    scheme = get_security_scheme(validator, scheme_name)
    if scheme is None:
        return False
    return cast(
        bool,
        scheme.type == SecuritySchemeType.HTTP
        and scheme.scheme == HttpAuthScheme.BEARER,
    )


def validate_security(
    validator: RequestValidator, request: OpenAPIRequest, operation: Operation
) -> MappingStrAny:
    """Validate security data for the request if any.

    First, check whether operation is secured or there is global security
    definitions. If not and current operation is not secured - return empty
    :class:`pyrsistent.PMap`.

    If operation secured, go through all security items and attempt to match
    their data in request. If some of security items is matched - return it
    as result, if not - raise `SecurityError`.
    """
    security_list = operation.security or validator.spec.security
    if not security_list:
        return pmap()

    # If operation "secured" with an empty object it means the whole security
    # rules are optional. However to not return empty security details it is
    # needed to move given security scheme to the bottom
    has_empty_security = {} in security_list
    if has_empty_security:
        security_list.remove({})
        security_list.append({})

    for item in security_list:
        data = {
            key: get_security_data(validator, request, key) for key in item
        }
        if all(value for value in data.values()):
            return pmap(data)

    # To supply proper security error need to check how many security schemas
    # should be applied for given operation
    #
    # If there is only one security schema with one security schema item and
    # it is a HTTP basic - raise BasicSecurityError (401 Unauthenticated). In
    # all other cases - raise a SecurityError (403 Access Denied)
    if len(security_list) == 1 and len(security_list[0]) == 1:
        scheme_name = list(security_list[0].keys())[0]
        if is_basic_auth_security_scheme(validator, scheme_name):
            raise BasicSecurityError()

    raise SecurityError()



from typing import cast, Optional, Union

from aiohttp import hdrs, web
from aiohttp.payload import IOBasePayload, Payload
from openapi_core.schema.operations.models import Operation
from openapi_core.schema.specs.models import Spec
from openapi_core.validation.request.datatypes import (
    OpenAPIRequest,
    RequestParameters,
)
from openapi_core.validation.response.datatypes import OpenAPIResponse
from yarl import URL

from .constants import HANDLER_OPENAPI_MAPPING_KEY
from .exceptions import OperationError
from .utils import get_openapi_spec
from ..annotations import Handler


def find_core_operation(
    request: web.Request, handler: Handler
) -> Optional[Operation]:
    mapping = getattr(handler, HANDLER_OPENAPI_MAPPING_KEY, None)
    if not mapping:
        return None

    operation_id = mapping.get(request.method) or mapping.get(hdrs.METH_ANY)
    if operation_id is None:
        return None

    try:
        return get_core_operation(
            get_openapi_spec(request.config_dict), operation_id
        )
    except OperationError:
        return None


def get_core_operation(spec: Spec, operation_id: str) -> Operation:
    for path in spec.paths.values():
        for operation in path.operations.values():
            if operation.operation_id == operation_id:
                return operation
    raise OperationError(
        f"Unable to find operation '{operation_id}' in given OpenAPI spec"
    )


def get_full_url_pattern(request: web.Request) -> str:
    """Get full URL pattern for given :class:`aiohttp.web.Request` instance."""
    full_url: URL = request.url.with_path(get_path_pattern(request))
    return full_url.human_repr()


def get_path_pattern(request: web.Request) -> str:
    """Get path pattern for given :class:`aiohttp.web.Request` instance.

    When current handler is a dynamic route: use formatter, otherwise use
    path from route info.
    """
    info = request.match_info.route.get_info()
    formatter = info.get("formatter")
    return cast(str, formatter if formatter is not None else info.get("path"))


async def to_core_openapi_request(request: web.Request) -> OpenAPIRequest:
    """Convert aiohttp.web request to openapi-core request.

    Afterwards opeanpi-core request can be used for validation request data
    against spec.
    """
    body: Optional[Union[bytes, str]] = None
    if request.body_exists and request.can_read_body:
        raw_body = await request.read()

        # If it possible, convert bytes to string
        try:
            body = raw_body.decode("utf-8")
        # If not, use bytes as request body instead
        except UnicodeDecodeError:
            body = raw_body

    return OpenAPIRequest(
        full_url_pattern=get_full_url_pattern(request),
        method=request.method.lower(),
        body=body,
        mimetype=request.content_type,
        parameters=to_core_request_parameters(request),
    )


def to_core_openapi_response(response: web.StreamResponse) -> OpenAPIResponse:
    """Convert aiohttp.web response to openapi-core response."""
    return OpenAPIResponse(
        data=to_core_openapi_response_data(response),
        status_code=response.status,
        mimetype=response.content_type,
    )


def to_core_openapi_response_data(
    response: web.StreamResponse,
) -> Optional[bytes]:
    if isinstance(response, web.Response):
        body = response.body
        if not body:
            return None

        # TODO: Find better way to provide response from payload
        if isinstance(body, IOBasePayload):
            return cast(bytes, body._value.getvalue())

        if isinstance(body, Payload):
            return cast(bytes, body._value)

        return body
    return None


def to_core_request_parameters(request: web.Request) -> RequestParameters:
    header_attr = [
        item
        for item in RequestParameters.__attrs_attrs__
        if item.name == "header"
    ][0]
    is_dict_factory = header_attr.default.factory == dict

    return RequestParameters(
        query=request.rel_url.query,
        header=request.headers if is_dict_factory else request.headers.items(),
        cookie=request.cookies,
        path=request.match_info,
    )



import inspect
import json
import os
import warnings
from functools import lru_cache, partial
from pathlib import Path
from typing import Callable, cast, Deque, List, overload, Tuple, Union

import attr
import yaml
from aiohttp import hdrs, web
from aiohttp_middlewares import cors_middleware
from openapi_core.schema.specs.models import Spec
from openapi_core.shortcuts import create_spec
from pyrsistent import pmap
from yarl import URL

from . import views
from .constants import (
    APP_OPENAPI_SCHEMA_KEY,
    APP_OPENAPI_SPEC_KEY,
    HANDLER_OPENAPI_MAPPING_KEY,
)
from .core_data import get_core_operation
from .exceptions import ConfigurationError
from .middlewares import openapi_middleware
from .utils import add_prefix
from ..annotations import (
    DictStrAny,
    DictStrStr,
    F,
    Handler,
    Protocol,
    ViewType,
)
from ..settings import APP_SETTINGS_KEY, BaseSettings


SchemaLoader = Callable[[bytes], DictStrAny]
Url = Union[str, URL]


class CreateSchemaAndSpec(Protocol):
    def __call__(
        self, path: Path, *, schema_loader: SchemaLoader = None
    ) -> Tuple[DictStrAny, Spec]:  # pragma: no cover
        ...


@attr.dataclass(slots=True)
class OperationTableDef:
    """Map OpenAPI 3 operations to aiohttp.web view handlers.

    In short it is rororo's equialent to :class:`aiohttp.web.RouteTableDef`.
    Under the hood, on :func:`rororo.openapi.setup_openapi` it
    still will use ``RouteTableDef`` for registering view handlers to
    :class:`aiohttp.web.Application`.

    But unlike ``RouteTableDef`` it does not register any HTTP method handlers
    (as via ``@routes.get`` decorator) in favor of just registering the
    operations.

    There are two ways for registering view hanlder,

    1. With bare ``@operations.register`` call when OpenAPI ``operationId``
       equals to view handler name.
    2. Or by passing ``operation_id`` to the decorator as first arg, when
       ``operationId`` does not match view handler name, or if you don't like
       the fact of guessing operation ID from view handler name.

    Both of ways described below,

    .. code-block:: python

        from rororo import OperationTableDef

        operations = OperationTableDef()

        # Expect OpenAPI 3 schema to contain operationId: hello_world
        @operations.register
        async def hello_world(request: web.Request) -> web.Response:
            ...


        # Explicitly use `operationId: "helloWorld"`
        @operations.register("helloWorld")
        async def hello_world(request: web.Request) -> web.Response:
            ...

    Class based views supported as well. In most generic way you just need
    to decorate your view with ``@operations.register`` decorator and ensure
    that ``operationId`` equals to view method qualified name
    (``__qualname__``).

    For example,

    .. code-block:: python

        @operations.register
        class UserView(web.View):
            async def get(self) -> web.Response:
                ...

    expects for operation ID ``UserView.get`` to be declared in OpenAPI schema.

    In same time,

    .. code-block:: python

        @operations.register("users")
        class UserView(web.View):
            async def get(self) -> web.Response:
                ...

    expects for operation ID ``users.get`` to be declared in OpenAPI schema.

    Finally,

    .. code-block:: python

        @operations.register
        class UserView(web.View):
            @operations.register("me")
            async def get(self) -> web.Response:
                ...

    expects for operation ID ``me`` to be declared in OpenAPI schema.

    When the class based view provides mutliple view methods (for example
    ``delete``, ``get``, ``patch`` & ``put``) *rororo* expects that
    OpenAPI schema contains operation IDs for each of view method.

    If supplied ``operation_id`` does not exist in OpenAPI 3 schema,
    :func:`rororo.openapi.setup_openapi` call raises an ``OperationError``.
    """

    handlers: List[Handler] = attr.Factory(list)
    views: List[ViewType] = attr.Factory(list)

    def __add__(self, other: "OperationTableDef") -> "OperationTableDef":
        return OperationTableDef(
            handlers=[*self.handlers, *other.handlers],
            views=[*self.views, *other.views],
        )

    def __iadd__(self, other: "OperationTableDef") -> "OperationTableDef":
        self.handlers.extend(other.handlers)
        self.views.extend(other.views)
        return self

    @overload
    def register(self, handler: F) -> F:
        ...  # pragma: no cover

    @overload
    def register(self, operation_id: str) -> Callable[[F], F]:
        ...  # pragma: no cover

    def register(self, mixed):  # type: ignore
        operation_id = mixed if isinstance(mixed, str) else mixed.__qualname__

        def decorator(handler: F) -> F:
            mapping: DictStrStr = {}

            if self._is_view(handler):
                mapping.update(
                    self._register_view(handler, operation_id)  # type: ignore
                )
            else:
                mapping.update(self._register_handler(handler, operation_id))

            setattr(handler, HANDLER_OPENAPI_MAPPING_KEY, pmap(mapping))
            return handler

        return decorator(mixed) if callable(mixed) else decorator

    def _is_view(self, handler: F) -> bool:
        is_class = inspect.isclass(handler)
        return is_class and issubclass(handler, web.View)  # type: ignore

    def _register_handler(
        self, handler: Handler, operation_id: str
    ) -> DictStrStr:
        # Hacky way to check whether handler is a view function or view method
        has_self_parameter = "self" in inspect.signature(handler).parameters

        # Register only view functions, view methods will be registered via
        # view class instead
        if not has_self_parameter:
            self.handlers.append(handler)

        return {hdrs.METH_ANY: operation_id}

    def _register_view(self, view: ViewType, prefix: str) -> DictStrStr:
        mapping: DictStrStr = {}

        for value in vars(view).values():
            if not callable(value):
                continue

            name = value.__name__
            maybe_method = name.upper()
            if maybe_method not in hdrs.METH_ALL:
                continue

            maybe_operation_id = getattr(
                value, HANDLER_OPENAPI_MAPPING_KEY, {}
            ).get(hdrs.METH_ANY)
            mapping[maybe_method] = (
                maybe_operation_id
                if maybe_operation_id
                else f"{prefix}.{name}"
            )

        self.views.append(view)
        return mapping


def convert_operations_to_routes(
    operations: OperationTableDef, spec: Spec, *, prefix: str = None
) -> web.RouteTableDef:
    """Convert operations table defintion to routes table definition."""

    async def noop(request: web.Request) -> web.Response:
        return web.json_response(status=204)  # pragma: no cover

    routes = web.RouteTableDef()

    # Add plain handlers to the route table def as a route
    for handler in operations.handlers:
        operation_id = getattr(handler, HANDLER_OPENAPI_MAPPING_KEY)[
            hdrs.METH_ANY
        ]
        core_operation = get_core_operation(spec, operation_id)

        routes.route(
            core_operation.http_method,
            add_prefix(core_operation.path_name, prefix),
            name=get_route_name(core_operation.operation_id),
        )(handler)

    # But view should be added as a view instead
    for view in operations.views:
        ids: Deque[str] = Deque(
            getattr(view, HANDLER_OPENAPI_MAPPING_KEY).values()
        )

        first_operation_id = ids.popleft()
        core_operation = get_core_operation(spec, first_operation_id)

        path = add_prefix(core_operation.path_name, prefix)
        routes.view(
            path,
            name=get_route_name(core_operation.operation_id),
        )(view)

        # Hacky way of adding aliases to class based views with multiple
        # registered view methods
        for other_operation_id in ids:
            routes.route(
                hdrs.METH_ANY, path, name=get_route_name(other_operation_id)
            )(noop)

    return routes


def create_schema_and_spec(
    path: Path, *, schema_loader: SchemaLoader = None
) -> Tuple[DictStrAny, Spec]:
    schema = read_openapi_schema(path, loader=schema_loader)
    return (schema, create_spec(schema))


@lru_cache(maxsize=128)
def create_schema_and_spec_with_cache(  # type: ignore
    path: Path, *, schema_loader: SchemaLoader = None
) -> Tuple[DictStrAny, Spec]:
    return create_schema_and_spec(path, schema_loader=schema_loader)


def find_route_prefix(
    oas: DictStrAny,
    *,
    server_url: Union[str, URL] = None,
    settings: BaseSettings = None,
) -> str:
    if server_url is not None:
        return get_route_prefix(server_url)

    servers = oas["servers"]
    if len(servers) == 1:
        return get_route_prefix(servers[0]["url"])

    if settings is None:
        raise ConfigurationError(
            "Unable to guess route prefix as OpenAPI schema contains "
            "multiple servers and aiohttp.web has no settings instance "
            "configured."
        )

    for server in servers:
        mixed = server.get("x-rororo-level")
        if isinstance(mixed, list):
            if settings.level in mixed:
                return get_route_prefix(server["url"])
        elif mixed == settings.level:
            return get_route_prefix(server["url"])

    raise ConfigurationError(
        "Unable to guess route prefix as no server in OpenAPI schema has "
        f'defined "x-rororo-level" key of "{settings.level}".'
    )


def get_route_name(operation_id: str) -> str:
    return operation_id.replace(" ", "-")


def get_route_prefix(mixed: Url) -> str:
    return (URL(mixed) if isinstance(mixed, str) else mixed).path


def read_openapi_schema(
    path: Path, *, loader: SchemaLoader = None
) -> DictStrAny:
    """Read OpenAPI Schema from given path.

    By default, when ``loader`` is not explicitly passed, attempt to guess
    schema loader function from path extension.

    ``loader`` should be a callable, which receives ``bytes`` and returns
    ``Dict[str, Any]`` of OpenAPI Schema.

    By default, next schema loader used,

    - :func:`json.loads` for ``openapi.json``
    - ``yaml.load`` for ``openapi.yaml``
    """
    if loader is None:
        if path.suffix == ".json":
            loader = json.loads
        elif path.suffix in {".yml", ".yaml"}:
            safe_loader = getattr(yaml, "CSafeLoader", yaml.SafeLoader)
            loader = partial(yaml.load, Loader=safe_loader)

    if loader is not None:
        return loader(path.read_bytes())

    raise ConfigurationError(
        f"Unsupported OpenAPI schema file: {path}. At a moment rororo "
        "supports loading OpenAPI schemas from: .json, .yml, .yaml files"
    )


@overload
def setup_openapi(
    app: web.Application,
    schema_path: Union[str, Path],
    *operations: OperationTableDef,
    server_url: Url = None,
    is_validate_response: bool = True,
    has_openapi_schema_handler: bool = True,
    use_error_middleware: bool = True,
    error_middleware_kwargs: DictStrAny = None,
    use_cors_middleware: bool = True,
    cors_middleware_kwargs: DictStrAny = None,
    schema_loader: SchemaLoader = None,
    cache_create_schema_and_spec: bool = False,
) -> web.Application:  # pragma: no cover
    ...


@overload
def setup_openapi(
    app: web.Application,
    *operations: OperationTableDef,
    schema: DictStrAny,
    spec: Spec,
    server_url: Url = None,
    is_validate_response: bool = True,
    has_openapi_schema_handler: bool = True,
    use_error_middleware: bool = True,
    error_middleware_kwargs: DictStrAny = None,
    use_cors_middleware: bool = True,
    cors_middleware_kwargs: DictStrAny = None,
) -> web.Application:  # pragma: no cover
    ...


def setup_openapi(  # type: ignore
    app: web.Application,
    schema_path: Union[str, Path] = None,
    *operations: OperationTableDef,
    schema: DictStrAny = None,
    spec: Spec = None,
    server_url: Url = None,
    is_validate_response: bool = True,
    has_openapi_schema_handler: bool = True,
    use_error_middleware: bool = True,
    error_middleware_kwargs: DictStrAny = None,
    use_cors_middleware: bool = True,
    cors_middleware_kwargs: DictStrAny = None,
    schema_loader: SchemaLoader = None,
    cache_create_schema_and_spec: bool = False,
) -> web.Application:
    """Setup OpenAPI schema to use with aiohttp.web application.

    Unlike `aiohttp-apispec <https://aiohttp-apispec.readthedocs.io/>`_ and
    other tools, which provides OpenAPI/Swagger support for aiohttp.web
    applications, *rororo* changes the way of using OpenAPI schema with
    ``aiohttp.web`` apps.

    *rororo* using schema first approach and relies on concrete OpenAPI schema
    file, path to which need to be registered on application startup (mostly
    inside of ``create_app`` factory or right after
    :class:`aiohttp.web.Application` instantiation).

    And as valid OpenAPI schema ensure unique ``operationId`` used accross the
    schema *rororo* uses them as a key while telling aiohttp.web to use given
    view handler for serving required operation.

    With that in mind registering (setting up) OpenAPI schema requires:

    1. :class:`aiohttp.web.Application` instance
    2. Path to file (json or yaml) with OpenAPI schema
    3. OpenAPI operation handlers mapping (rororo's equialent of
       :class:`aiohttp.web.RouteTableDef`)

    In common cases setup looks like,

    .. code-block:: python

        from pathlib import Path
        from typing import List

        from aiohttp import web

        from .views import operations


        def create_app(argv: List[str] = None) -> web.Application:
            return setup_openapi(
                web.Application(),
                Path(__file__).parent / "openapi.yaml",
                operations,
            )

    If your OpenAPI schema contains multiple servers schemas, like,

    .. code-block:: yaml

        servers:
        - url: "/api/"
          description: "Test environment"
        - url: "http://localhost:8080/api/"
          description: "Dev environment"
        - url: "http://prod.url/api/"
          description: "Prod environment"

    you have 2 options of telling *rororo* how to use specific server URL.

    First, is passing ``server_url``, while setting up OpenAPI, for example,

    .. code-block:: python

        setup_openapi(
            web.Application(),
            Path(__file__).parent / "openapi.yaml",
            operations,
            server_url=URL("http://prod.url/api/"),
        )

    Second, is more complicated as you need to wrap ``aiohttp.web`` application
    into :func:`rororo.settings.setup_settings` and mark each server with
    ``x-rororo-level`` special key in server schema definition as,

    .. code-block:: yaml

        servers:
        - url: "/api/"
          x-rororo-level: "test"
        - url: "http://localhost:8080/api/"
          x-rororo-level: "dev"
        - url: "http://prod.url/api/"
          x-rororo-level: "prod"

    After, *rororo* will try to equal current app settings level with the
    schema and if URL matched, will use given server URL for finding out
    route prefix.

    By default, *rororo* will validate operation responses against OpenAPI
    schema. To disable this feature, pass ``is_validate_response`` falsy flag.

    By default, *rororo* will share the OpenAPI schema which is registered
    for your aiohttp.web application. In case if you don't want to share this
    schema, pass ``has_openapi_schema_handler=False`` on setting up OpenAPI.

    By default, *rororo* will enable
    :func:`aiohttp_middlewares.cors.cors_middleware` without any settings and
    :func:`aiohttp_middlewares.error.error_middleware` with custom error
    handler to ensure that security / validation errors does not provide any
    mess to stdout. Pass ``use_cors_middleware`` /
    ``use_error_middleware`` to change or entirely disable this default
    behaviour.

    For passing custom options to CORS middleware, use
    ``cors_middleware_kwargs`` mapping. If kwarg does not support by CORS
    middleware - *rororo* will raise a ``ConfigurationError``. All list of
    options available at documentation for
    :func:`aiohttp_middlewares.cors.cors_middleware`.

    To simplify things *rororo* expects on OpenAPI 3 path and do reading schema
    from file and specifying ``openapi_core.schema.specs.models.Spec`` instance
    inside of :func:`rororo.openapi.setup_openapi` call.

    However, it is possible to completely customize this default behaviour and
    pass OpenAPI ``schema`` and ``spec`` instance directly. In that case
    ``schema`` keyword argument should contains raw OpenAPI 3 schema as
    ``Dict[str, Any]``, while ``spec`` to be an
    ``openapi_core.schema.specs.models.Spec`` instance.

    This behaviour might be helpful if you'd like to cache reading schema and
    instantiating spec within tests or other environments, which requires
    multiple :func:`rororo.openapi.setup_openapi` calls.

    .. code-block:: python

        from pathlib import Path

        import yaml
        from aiohttp import web
        from openapi_core.shortcuts import create_spec
        from rororo import setup_openapi


        # Reusable OpenAPI data
        openapi_yaml = Path(__file__).parent / "openapi.yaml"
        schema = yaml.load(
            openapi_yaml.read_bytes(), Loader=yaml.CSafeLoader
        )
        spec = create_spec(schema)

        # Create OpenAPI 3 aiohttp.web server application
        app = setup_openapi(web.Application(), schema=schema, spec=spec)

    For default behaviour, with passing ``schema_path``, there are few options
    on customizing schema load process as well,

    By default, *rororo* will use :func:`json.loads` to load OpenAPI schema
    content from JSON file and ``yaml.CSafeLoader`` if it is available to load
    schema content from YAML files (with fallback to ``yaml.SafeLoader``). But,
    for performance considreations, you might use any other function to load
    the schema. Example below illustrates how to use ``ujson.loads`` function
    to load content from JSON schema,

    .. code-block:: python

        import ujson

        app = setup_openapi(
            web.Application(),
            Path(__file__).parent / "openapi.json",
            operations,
            schema_loader=ujson.loads,
        )

    Schema loader function expects ``bytes`` as only argument and should return
    ``Dict[str, Any]`` as OpenAPI schema dict.

    .. danger::
        By default *rororo* does not cache slow calls to read OpenAPI schema
        and creating its spec. But sometimes, for example in tests, it is
        sufficient to cache those calls. To enable cache behaviour pass
        ``cache_create_schema_and_spec=True`` or even better,
        ``cache_create_schema_and_spec=settings.is_test``.

        But this may result in unexpected issues, as schema and spec will be
        cached once and on next call it will result cached data instead to
        attempt read fresh schema from the disk and instantiate OpenAPI Spec
        instance.
    """

    if isinstance(schema_path, OperationTableDef):
        operations = (schema_path, *operations)
        schema_path = None

    if schema is None and spec is None:
        if schema_path is None:
            raise ConfigurationError(
                "Please supply only `spec` keyword argument, or only "
                "`schema_path` positional argumnet, not both."
            )

        # Ensure OpenAPI schema is a readable file
        path = (
            Path(schema_path) if isinstance(schema_path, str) else schema_path
        )
        if not path.exists() or not path.is_file():
            uid = os.getuid()
            raise ConfigurationError(
                f"Unable to find OpenAPI schema file at {path}. Please check "
                "that file exists at given path and readable by current user "
                f"ID: {uid}"
            )

        # Create the spec and put it to the application dict as well
        create_func: CreateSchemaAndSpec = (
            create_schema_and_spec_with_cache  # type: ignore
            if cache_create_schema_and_spec
            else create_schema_and_spec
        )

        try:
            schema, spec = create_func(path, schema_loader=schema_loader)
        except Exception:
            raise ConfigurationError(
                f"Unable to load valid OpenAPI schema in {path}. In most "
                "cases it means that given file doesn't contain valid OpenAPI "
                "3 schema. To get full details about errors run "
                f"`openapi-spec-validator {path.absolute()}`"
            )
    elif schema_path is not None:
        warnings.warn(
            "You supplied `schema_path` positional argument as well as "
            "supplying `schema` & `spec` keyword arguments. `schema_path` "
            "will be ignored in favor of `schema` & `spec` args."
        )

    # Store schema and spec in application dict
    app[APP_OPENAPI_SCHEMA_KEY] = schema
    app[APP_OPENAPI_SPEC_KEY] = spec

    # Register the route to dump openapi schema used for the application if
    # required
    route_prefix = find_route_prefix(
        cast(DictStrAny, schema),
        server_url=server_url,
        settings=app.get(APP_SETTINGS_KEY),
    )
    if has_openapi_schema_handler:
        app.router.add_get(
            add_prefix("/openapi.{schema_format}", route_prefix),
            views.openapi_schema,
        )

    # Register all operation handlers to web application
    for item in operations:
        app.router.add_routes(
            convert_operations_to_routes(item, spec, prefix=route_prefix)
        )

    # Add OpenAPI middleware
    kwargs = error_middleware_kwargs or {}
    kwargs.setdefault("default_handler", views.default_error_handler)

    try:
        app.middlewares.insert(
            0,
            openapi_middleware(
                is_validate_response=is_validate_response,
                use_error_middleware=use_error_middleware,
                error_middleware_kwargs=kwargs,
            ),
        )
    except TypeError:
        raise ConfigurationError(
            "Unsupported kwargs passed to error middleware. Please check "
            "given kwargs and remove unsupported ones: "
            f"{error_middleware_kwargs!r}"
        )

    # Add CORS middleware if necessary
    if use_cors_middleware:
        try:
            app.middlewares.insert(
                0, cors_middleware(**(cors_middleware_kwargs or {}))
            )
        except TypeError:
            raise ConfigurationError(
                "Unsupported kwargs passed to CORS middleware. Please check "
                "given kwargs and remove unsupported ones: "
                f"{cors_middleware_kwargs!r}"
            )

    return app



from functools import partial

from aiohttp import web
from aiohttp_middlewares import error_middleware, get_error_response
from aiohttp_middlewares.annotations import Middleware

from .constants import REQUEST_CORE_OPERATION_KEY
from .core_data import find_core_operation
from .validators import validate_request, validate_response
from ..annotations import DictStrAny, Handler


def get_actual_handler(handler: Handler) -> Handler:
    """Remove partially applied middlewares from actual handler.

    aiohttp wraps handler into middlewares, so if any middlewares declared in
    application after ``openapi_middleware`` the handler will look like::

        functools.partial(
            <function middleware1.<locals>.middleware at 0x111325b80>,
            handler=functools.partial(
                <function middleware2.<locals>.middleware at 0x111325ca0>,
                handler=<function actual_handler at 0x1112aa700>
            )
        )

    In that case ``HANDLER_OPENAPI_MAPPING_KEY`` will not be accessed in the
    partial, which results that handler will not be validated against
    OpenAPI schema.
    """
    if isinstance(handler, partial) and "handler" in handler.keywords:
        return get_actual_handler(handler.keywords["handler"])
    return handler


def openapi_middleware(
    *,
    is_validate_response: bool = True,
    use_error_middleware: bool = True,
    error_middleware_kwargs: DictStrAny = None
) -> Middleware:
    """Middleware to handle requests to handlers covered by OpenAPI schema.

    In most cases you don't need to add it to list of ``web.Application``
    middlewares as :func:`rororo.openapi.setup_openapi` will setup it for you,
    but if, for some reason, you don't want to call high order
    ``setup_openapi`` function, you'll need to add given middleware to your
    :class:`aiohttp.web.Applicaiton` manually.
    """

    error_middleware_instance = (
        error_middleware(**error_middleware_kwargs or {})
        if use_error_middleware
        else None
    )

    async def get_response(
        request: web.Request, handler: Handler
    ) -> web.StreamResponse:
        if error_middleware_instance is None:
            return await handler(request)
        return await error_middleware_instance(request, handler)

    @web.middleware
    async def middleware(
        request: web.Request, handler: Handler
    ) -> web.StreamResponse:
        # At first, check that given handler registered as OpenAPI operation
        # handler. For this check remove all partially applied middlewares
        # from the handler,
        core_operation = find_core_operation(
            request, get_actual_handler(handler)
        )
        if core_operation is None:
            return await get_response(request, handler)

        try:
            # Run actual `aiohttp.web` handler for requested operation
            request[REQUEST_CORE_OPERATION_KEY] = core_operation
            response = await get_response(
                await validate_request(request), handler
            )

            # For performance considerations it is useful to turn off
            # validating responses at production environment as unfortunately
            # it will need to do extra checks after response is ready
            if is_validate_response:
                validate_response(request, response)

            return response
        except Exception as err:
            return await get_error_response(
                request, err, **error_middleware_kwargs or {}
            )

    return middleware



import logging

import yaml
from aiohttp import web
from aiohttp_middlewares import error_context

from .exceptions import ConfigurationError, OpenAPIError
from .utils import get_openapi_schema
from ..annotations import MappingStrStr


logger = logging.getLogger(__name__)


async def default_error_handler(request: web.Request) -> web.Response:
    """Default error handler which will ignore logging OpenAPI errors."""
    with error_context(request) as context:
        err = context.err
        headers: MappingStrStr = {}

        if isinstance(err, OpenAPIError):
            headers = err.headers
        else:
            logger.error(context.message, exc_info=True)

        return web.json_response(
            context.data,
            status=context.status,
            headers=headers,
        )


async def openapi_schema(request: web.Request) -> web.Response:
    """Dump OpenAPI Schema into specified format."""
    schema_format = request.match_info.get("schema_format")
    schema = get_openapi_schema(request.config_dict)

    if schema_format == "json":
        return web.json_response(schema)

    if schema_format == "yaml":
        safe_dumper = getattr(yaml, "CSafeDumper", yaml.SafeDumper)
        return web.Response(
            text=yaml.dump(schema, Dumper=safe_dumper),
            content_type="application/yaml",
        )

    raise ConfigurationError(
        f"Schema format {schema_format} not supported at a moment."
    )



"""
======
rororo
======

`OpenAPI 3 <https://spec.openapis.org/oas/v3.0.2>`_ schema support
for `aiohttp.web <https://aiohttp.readthedocs.io/en/stable/web.html>`_
applications.

As well as bunch other utilities to build effective web applications with
Python 3 & ``aiohttp.web``.

"""

from .openapi import (
    get_openapi_context,
    get_openapi_schema,
    get_openapi_spec,
    get_validated_data,
    openapi_context,
    OperationTableDef,
    setup_openapi,
)
from .settings import BaseSettings, setup_settings, setup_settings_from_environ


__all__ = (
    "BaseSettings",
    "get_openapi_context",
    "get_openapi_schema",
    "get_openapi_spec",
    "get_validated_data",
    "openapi_context",
    "OperationTableDef",
    "setup_openapi",
    "setup_settings",
    "setup_settings_from_environ",
)

__author__ = "Igor Davydenko"
__license__ = "BSD-3-Clause"
__version__ = "2.0.2"



"""
==================
rororo.annotations
==================

Internal module to keep all reusable type annotations for the project in one
place.

"""

import types
from typing import Any, Callable, Dict, Mapping, Type, TypeVar, Union

try:
    from typing_extensions import Literal, Protocol, TypedDict
except ImportError:
    from typing import Literal, Protocol, TypedDict  # type: ignore

from aiohttp import web
from aiohttp.web_middlewares import _Handler


F = TypeVar("F", bound=Callable[..., Any])  # noqa: VNE001
T = TypeVar("T")

Level = Literal["test", "dev", "staging", "prod"]

Handler = _Handler
ViewType = Type[web.View]

DictStrAny = Dict[str, Any]
DictStrInt = Dict[str, int]
DictStrStr = Dict[str, str]

MappingStrAny = Mapping[str, Any]
MappingStrInt = Mapping[str, int]
MappingStrStr = Mapping[str, str]

Settings = Union[types.ModuleType, DictStrAny]


(Protocol, TypedDict)  # Make flake8 happy



"""
===================
rororo.openapi.data
===================

Provide structures for OpenAPI data.

"""

from typing import Any

import attr
from aiohttp import web
from aiohttp.helpers import ChainMapProxy
from openapi_core.validation.request.datatypes import RequestParameters
from pyrsistent import pmap

from ..annotations import MappingStrAny


@attr.dataclass(frozen=True, slots=True)
class OpenAPIParameters:
    #: Path parameters
    path: MappingStrAny = attr.Factory(pmap)

    #: Query parameters
    query: MappingStrAny = attr.Factory(pmap)

    #: Parameters from headers
    header: MappingStrAny = attr.Factory(pmap)

    #: Cookie parameters
    cookie: MappingStrAny = attr.Factory(pmap)


@attr.dataclass(frozen=True, slots=True)
class OpenAPIContext:
    """All data associated with current request to OpenAPI handler.

    Contains only valid parameters, security data & request body data.
    Example bellow illustrates how to work with context data,

    .. code-block:: python

        from rororo import get_openapi_context
        from rororo.openapi.exceptions import InvalidCredentials


        async def create_user(request: web.Request) -> web.Response:
            context = get_openapi_context(request)

            # Authenticate current user (accessing security data)
            if not authenticate(api_key=context.security["apiKey"]):
                raise InvalidCredentials()

            # Add new user (accessing request body data)
            async with request.config_dict["db"].acquire() as conn:
                user = await create_user(
                    conn,
                    email=context.data["email"],
                    password=context.data["password"],
                )

            # Return response due to query string param
            # (accessing parameters data)
            if context.parameters.query["login"]:
                return web.json_response(
                    request.app.router["login"].url_for()
                )

            return web.json_response(user.to_api_dict())

    """

    #: Request instance
    request: web.Request

    #: Application instance
    app: web.Application

    #: Config dict instance
    config_dict: ChainMapProxy

    #: Request parameters instance
    parameters: OpenAPIParameters = attr.Factory(OpenAPIParameters)

    #: Request security data
    security: MappingStrAny = attr.Factory(pmap)

    #: Request body data
    data: Any = None


def to_openapi_parameters(
    core_parameters: RequestParameters,
) -> OpenAPIParameters:
    """Convert openapi-core parameters to internal parameters instance."""
    return OpenAPIParameters(
        path=pmap(core_parameters.path),
        query=pmap(core_parameters.query),
        header=pmap(core_parameters.header),
        cookie=pmap(core_parameters.cookie),
    )



from contextlib import contextmanager
from typing import Iterator

from aiohttp import web

from .data import OpenAPIContext
from .utils import get_openapi_context


@contextmanager
def openapi_context(request: web.Request) -> Iterator[OpenAPIContext]:
    """Context manager to access valid OpenAPI data for given request.

    If request validation done well and request to OpenAPI operation view
    handler is valid one, view handler may need to use request data for its
    needs. To achieve it use given context manager as,

    .. code-block:: python

        from rororo import openapi_context, OperationTableDef

        operations = OperationTableDef()


        @operations.register
        async def hello_world(request: web.Request) -> web.Response:
            with openapi_context(request) as context:
                ...

    If using context managers inside of view handlers considered as unwanted,
    there is an other option in
    :func:`rororo.openapi.get_openapi_context` function.
    """
    yield get_openapi_context(request)



"""
================
rororo.timedelta
================

Useful functions to work with timedelta instances.

"""

import datetime
import re
from typing import Optional

from .annotations import DictStrInt
from .utils import to_int


__all__ = (
    "str_to_timedelta",
    "timedelta_average",
    "timedelta_div",
    "timedelta_seconds",
    "timedelta_to_str",
)


SECONDS_PER_DAY = 86400
SECONDS_PER_WEEK = 604800
TIMEDELTA_FORMAT = "G:i"
TIMEDELTA_FORMATS = {
    "d": ("%(days)02d", r"(?P<days>\d{2,})"),
    "f": (
        "%(weeks)d%(short_weeks_label)s "
        "%(week_days)d%(short_week_days_label)s "
        "%(day_hours)d:%(hour_minutes)02d:%(minute_seconds)02d",
        r"((?P<weeks>\d+)%(short_weeks_label)s )?"
        r"((?P<week_days>\d{1,})%(short_week_days_label)s )?"
        r"(?P<hours>\d{1,2})\:(?P<minutes>\d{2})(\:(?P<seconds>\d{2}))?",
    ),
    "F": (
        "%(weeks)d %(weeks_label)s, %(week_days)d %(week_days_label)s, "
        "%(day_hours)d:%(hour_minutes)02d:%(minute_seconds)02d",
        r"((?P<weeks>\d+) %(weeks_label)s, )?"
        r"((?P<week_days>\d{1,}) %(week_days_label)s, )?"
        r"(?P<hours>\d{1,2})\:(?P<minutes>\d{2})(\:(?P<seconds>\d{2}))?",
    ),
    "g": ("%(day_hours)d", r"(?P<day_hours>\d{1,2})"),
    "G": ("%(hours)d", r"(?P<hours>\d+)"),
    "h": ("%(day_hours)02d", r"(?P<day_hours>\d{2})"),
    "H": ("%(hours)02d", r"(?P<hours>\d{2,})"),
    "i": ("%(hour_minutes)02d", r"(?P<hour_minutes>\d{2})"),
    "I": ("%(minutes)02d", r"(?P<minutes>\d{2,})"),
    "j": ("%(days)d", r"(?P<days>\d+)"),
    "l": ("%(days_label)s", r"%(days_label)s"),
    "L": ("%(weeks_label)s", r"%(weeks_label)s"),
    "m": ("%(week_days_label)s", r"%(week_days_label)s"),
    "r": (
        "%(days)d%(short_days_label)s "
        "%(day_hours)d:%(hour_minutes)02d:%(minute_seconds)02d",
        r"((?P<days>\d+)%(short_days_label)s )?"
        r"(?P<hours>\d{1,2})\:(?P<minutes>\d{2})\:(?P<seconds>\d{2})",
    ),
    "R": (
        "%(days)d %(days_label)s, "
        "%(day_hours)d:%(hour_minutes)02d:%(minute_seconds)02d",
        r"((?P<days>\d+) %(days_label)s, )?"
        r"(?P<hours>\d{1,2})\:(?P<minutes>\d{2})\:(?P<seconds>\d{2})",
    ),
    "s": ("%(minute_seconds)02d", r"(?P<minute_seconds>\d{2})"),
    "S": ("%(seconds)02d", r"(?P<seconds>{2,})"),
    "": ("%(microseconds)d", r"(?P<microseconds>\d{1,})"),
    "w": ("%(week_days)d", r"(?P<week_days>\d{1})"),
    "W": ("%(weeks)d", r"(?P<weeks>\d+)"),
}


def str_to_timedelta(
    value: str, fmt: str = None
) -> Optional[datetime.timedelta]:
    """
    Convert string value to timedelta instance according to the given format.

    If format not set function tries to load timedelta using default
    ``TIMEDELTA_FORMAT`` and then both of magic "full" formats.

    You should also specify list of formats and function tries to convert
    to timedelta using each of formats in list. First matched format would
    return the converted timedelta instance.

    If user specified format, but function cannot convert string to
    new timedelta instance - ``ValueError`` would be raised. But if user did
    not specify the format, function would be fail silently and return ``None``
    as result.

    :param value: String representation of timedelta.
    :param fmt: Format to use for conversion.
    """

    def timedelta_kwargs(data: DictStrInt) -> DictStrInt:
        """
        Convert day_hours, hour_minutes, minute_seconds, week_days and weeks to
        timedelta seconds.
        """
        seconds = data.get("seconds", 0)
        seconds += data.get("day_hours", 0) * 3600
        seconds += data.pop("hour_minutes", 0) * 60
        seconds += data.pop("minute_seconds", 0)
        seconds += data.pop("week_days", 0) * SECONDS_PER_DAY
        seconds += data.pop("weeks", 0) * SECONDS_PER_WEEK
        data.update({"seconds": seconds})
        return data

    if not isinstance(value, str):
        raise ValueError(
            'Value should be a "str" instance. You use {0}.'.format(
                type(value)
            )
        )

    user_fmt = fmt

    if isinstance(fmt, (list, tuple)):
        formats = list(fmt)
    elif fmt is None:
        formats = [TIMEDELTA_FORMAT, "F", "f"]
    else:
        formats = [fmt]

    locale_data = {
        "days_label": "({0}|{1})".format("day", "days"),
        "short_days_label": "d",
        "short_week_days_label": "d",
        "short_weeks_label": "w",
        "week_days_label": "({0}|{1})".format("day", "days"),
        "weeks_label": "({0}|{1})".format("week", "weeks"),
    }
    regexps = []

    for item in formats:
        processed = r"^"

        for part in item:
            if part in TIMEDELTA_FORMATS:
                part = TIMEDELTA_FORMATS[part][1] % locale_data
            else:
                part = re.escape(part)
            processed += part

        processed += r"$"
        regexps.append(processed)

    for regexp in regexps:
        timedelta_re = re.compile(regexp)
        matched = timedelta_re.match(value)

        if matched:
            data = {
                key: to_int(value) or 0
                for key, value in matched.groupdict().items()
            }
            return datetime.timedelta(**timedelta_kwargs(data))

    if user_fmt:
        err = "Cannot convert {0!r} to timedelta instance, using {1!r} format."
        raise ValueError(err.format(value, user_fmt))

    return None


def timedelta_average(*values: datetime.timedelta) -> datetime.timedelta:
    r"""Compute the arithmetic mean for timedeltas list.

    :param \*values: Timedelta instances to process.
    """
    if isinstance(values[0], (list, tuple)):
        values = values[0]
    return sum(values, datetime.timedelta()) // len(values)


def timedelta_div(
    first: datetime.timedelta, second: datetime.timedelta
) -> Optional[float]:
    """Implement divison for timedelta instances.

    :param first: First timedelta instance.
    :param second: Second timedelta instance.
    """
    first_seconds = timedelta_seconds(first)
    second_seconds = timedelta_seconds(second)

    if not second_seconds:
        return None

    return first_seconds / second_seconds


def timedelta_seconds(value: datetime.timedelta) -> int:
    """Return full number of seconds from timedelta.

    By default, Python returns only one day seconds, not all timedelta seconds.

    :param value: Timedelta instance.
    """
    return SECONDS_PER_DAY * value.days + value.seconds


def timedelta_to_str(value: datetime.timedelta, fmt: str = None) -> str:
    """Display the timedelta formatted according to the given string.

    You should use global setting ``TIMEDELTA_FORMAT`` to specify default
    format to this function there (like ``DATE_FORMAT`` for builtin ``date``
    template filter).

    Default value for ``TIMEDELTA_FORMAT`` is ``'G:i'``.

    Format uses the same policy as Django ``date`` template filter or
    PHP ``date`` function with several differences.

    Available format strings:

    +------------------+-----------------------------+------------------------+
    | Format character | Description                 | Example output         |
    +==================+=============================+========================+
    | ``a``            | Not implemented.            |                        |
    +------------------+-----------------------------+------------------------+
    | ``A``            | Not implemented.            |                        |
    +------------------+-----------------------------+------------------------+
    | ``b``            | Not implemented.            |                        |
    +------------------+-----------------------------+------------------------+
    | ``B``            | Not implemented.            |                        |
    +------------------+-----------------------------+------------------------+
    | ``c``            | Not implemented.            |                        |
    +------------------+-----------------------------+------------------------+
    | ``d``            | Total days, 2 digits with   | ``'01'``, ``'41'``     |
    |                  | leading zeros. Do not       |                        |
    |                  | combine with ``w`` format.  |                        |
    +------------------+-----------------------------+------------------------+
    | ``D``            | Not implemented.            |                        |
    +------------------+-----------------------------+------------------------+
    | ``f``            | Magic "full" format with    | ``'2w 4d 1:28:07'``    |
    |                  | short labels.               |                        |
    +------------------+-----------------------------+------------------------+
    | ``F``            | Magic "full" format with    | ``'2 weeks, 4 days,    |
    |                  | normal labels.              | 1:28:07'``             |
    +------------------+-----------------------------+------------------------+
    | ``g``            | Day, not total, hours       | ``'0'`` to ``'23'``    |
    |                  | without leading zeros. To   |                        |
    |                  | use with ``d``, ``j``, or   |                        |
    |                  | ``w``.                      |                        |
    +------------------+-----------------------------+------------------------+
    | ``G``            | Total hours without         | ``'1'``, ``'433'``     |
    |                  | leading zeros. Do not       |                        |
    |                  | combine with ``g`` or       |                        |
    |                  | ``h`` formats.              |                        |
    +------------------+-----------------------------+------------------------+
    | ``h``            | Day, not total, hours with  | ``'00'`` to ``'23'``   |
    |                  | leading zeros. To use with  |                        |
    |                  | ``d`` or ``w``.             |                        |
    +------------------+-----------------------------+------------------------+
    | ``H``            | Total hours with leading    | ``'01', ``'433'``      |
    |                  | zeros. Do not combine with  |                        |
    |                  | ``g`` or ``h`` formats.     |                        |
    +------------------+-----------------------------+------------------------+
    | ``i``            | Hour, not total, minutes, 2 | ``00`` to ``'59'``     |
    |                  | digits with leading zeros   |                        |
    |                  | To use with ``g``, ``G``,   |                        |
    |                  | ``h`` or ``H`` formats.     |                        |
    +------------------+-----------------------------+------------------------+
    | ``I``            | Total minutes, 2 digits or  | ``'01'``, ``'433'``    |
    |                  | more with leading zeros. Do |                        |
    |                  | not combine with ``i``      |                        |
    |                  | format.                     |                        |
    +------------------+-----------------------------+------------------------+
    | ``j``            | Total days, one or 2 digits | ``'1'``, ``'41'``      |
    |                  | without leading zeros. Do   |                        |
    |                  | not combine with ``w``      |                        |
    |                  | format.                     |                        |
    +------------------+-----------------------------+------------------------+
    | ``J``            | Not implemented.            |                        |
    +------------------+-----------------------------+------------------------+
    | ``l``            | Days long label.            | ``'day'`` or           |
    |                  | Pluralized and localized.   | ``'days'``             |
    +------------------+-----------------------------+------------------------+
    | ``L``            | Weeks long label.           | ``'week'`` or          |
    |                  | Pluralized and localized.   | ``'weeks'``            |
    +------------------+-----------------------------+------------------------+
    | ``m``            | Week days long label.       | ``'day'`` or           |
    |                  | Pluralized and localized.   | ``'days'``             |
    +------------------+-----------------------------+------------------------+
    | ``M``            | Not implemented.            |                        |
    +------------------+-----------------------------+------------------------+
    | ``n``            | Not implemented.            |                        |
    +------------------+-----------------------------+------------------------+
    | ``N``            | Not implemented.            |                        |
    +------------------+-----------------------------+------------------------+
    | ``O``            | Not implemented.            |                        |
    +------------------+-----------------------------+------------------------+
    | ``P``            | Not implemented.            |                        |
    +------------------+-----------------------------+------------------------+
    | ``r``            | Standart Python timedelta   | ``'18 d 1:28:07'``     |
    |                  | representation with short   |                        |
    |                  | labels.                     |                        |
    +------------------+-----------------------------+------------------------+
    | ``R``            | Standart Python timedelta   | ``'18 days, 1:28:07'`` |
    |                  | representation with normal  |                        |
    |                  | labels.                     |                        |
    +------------------+-----------------------------+------------------------+
    | ``s``            | Minute, not total, seconds, | ``'00'`` to ``'59'``   |
    |                  | 2 digits with leading       |                        |
    |                  | zeros. To use with ``i`` or |                        |
    |                  | ``I``.                      |                        |
    +------------------+-----------------------------+------------------------+
    | ``S``            | Total seconds. 2 digits or  | ``'00'``, ``'433'``    |
    |                  | more with leading zeros. Do |                        |
    |                  | not combine with ``s``      |                        |
    |                  | format.                     |                        |
    +------------------+-----------------------------+------------------------+
    | ``t``            | Not implemented.            |                        |
    +------------------+-----------------------------+------------------------+
    | ``T``            | Not implemented.            |                        |
    +------------------+-----------------------------+------------------------+
    | ``u``            | Second, not total,          | ``0`` to ``999999``    |
    |                  | microseconds.               |                        |
    +------------------+-----------------------------+------------------------+
    | ``U``            | Not implemented.            |                        |
    +------------------+-----------------------------+------------------------+
    | ``w``            | Week, not total, days, one  | ``0`` to ``6``         |
    |                  | digit without leading       |                        |
    |                  | zeros. To use with ``W``.   |                        |
    +------------------+-----------------------------+------------------------+
    | ``W``            | Total weeks, one or more    | ``'1'``, ``'41'``      |
    |                  | digits without leading      |                        |
    |                  | zeros.                      |                        |
    +------------------+-----------------------------+------------------------+
    | ``y``            | Not implemented.            |                        |
    +------------------+-----------------------------+------------------------+
    | ``Y``            | Not implemented.            |                        |
    +------------------+-----------------------------+------------------------+
    | ``z``            | Not implemented.            |                        |
    +------------------+-----------------------------+------------------------+
    | ``Z``            | Not implemented.            |                        |
    +------------------+-----------------------------+------------------------+

    For example,

    ::

        >>> import datetime
        >>> from rororo.timedelta import timedelta_to_str
        >>> delta = datetime.timedelta(seconds=99660)
        >>> timedelta_to_str(delta)
        ... '27:41'
        >>> timedelta_to_str(delta, 'r')
        ... '1d 3:41:00'
        >>> timedelta_to_str(delta, 'f')
        ... '1d 3:41'
        >>> timedelta_to_str(delta, 'W L, w l, H:i:s')
        ... '0 weeks, 1 day, 03:41:00'

    Couple words about magic "full" formats. These formats show weeks number
    with week label, days number with day label and seconds only if weeks
    number, days number or seconds greater that zero.

    For example,

    ::

        >>> import datetime
        >>> from rororo.timedelta import timedelta_to_str
        >>> delta = datetime.timedelta(hours=12)
        >>> timedelta_to_str(delta, 'f')
        ... '12:00'
        >>> timedelta_to_str(delta, 'F')
        ... '12:00'
        >>> delta = datetime.timedelta(hours=12, seconds=30)
        >>> timedelta_to_str(delta, 'f')
        ... '12:00:30'
        >>> timedelta_to_str(delta, 'F')
        ... '12:00:30'
        >>> delta = datetime.timedelta(hours=168)
        >>> timedelta_to_str(delta, 'f')
        ... '1w 0:00'
        >>> timedelta_to_str(delta, 'F')
        ... '1 week, 0:00'

    :param value: Timedelta instance to convert to string.
    :param fmt: Format to use for conversion.
    """
    # Only ``datetime.timedelta`` instances allowed for this function
    if not isinstance(value, datetime.timedelta):
        err = 'Value should be a "datetime.timedelta" instance. You use {0}.'
        raise ValueError(err.format(type(value)))

    # Generate total data
    days = value.days
    microseconds = value.microseconds
    seconds = timedelta_seconds(value)

    hours = seconds // 3600
    minutes = seconds // 60
    weeks = days // 7

    # Generate collapsed data
    day_hours = hours - days * 24
    hour_minutes = minutes - hours * 60
    minute_seconds = seconds - minutes * 60
    week_days = days - weeks * 7

    days_label = "day" if days % 10 == 1 else "days"
    short_days_label = "d"
    short_week_days_label = "d"
    short_weeks_label = "w"
    week_days_label = "day" if week_days % 10 == 1 else "days"
    weeks_label = "week" if weeks % 10 == 1 else "weeks"

    # Collect data
    data = locals()

    fmt = fmt or TIMEDELTA_FORMAT
    processed = ""

    for part in fmt:
        if part in TIMEDELTA_FORMATS:
            is_full_part = part in ("f", "F")
            is_repr_part = part in ("r", "R")

            part = TIMEDELTA_FORMATS[part][0]

            if is_full_part or is_repr_part:
                if is_repr_part and not days:
                    part = part.replace("%(days)d", "")
                    part = part.replace("%(days_label)s,", "")
                    part = part.replace("%(short_days_label)s", "")

                if is_full_part and not minute_seconds:
                    part = part.replace(":%(minute_seconds)02d", "")

                if is_full_part and not weeks:
                    part = part.replace("%(weeks)d", "")
                    part = part.replace("%(short_weeks_label)s", "")
                    part = part.replace("%(weeks_label)s,", "")

                if is_full_part and not week_days:
                    part = part.replace("%(week_days)d", "")
                    part = part.replace("%(short_week_days_label)s", "")
                    part = part.replace("%(week_days_label)s,", "")

                part = part.strip()
                part = " ".join(part.split())

        processed += part

    return processed % data



from collections import deque
from functools import partial
from typing import Any, cast, Dict, Iterator, List, Optional, Tuple

import pyrsistent
from email_validator import EmailNotValidError, validate_email
from isodate import parse_datetime
from jsonschema.exceptions import FormatError
from more_itertools import peekable
from openapi_core.casting.schemas.exceptions import CastError as CoreCastError
from openapi_core.exceptions import OpenAPIError as CoreOpenAPIError
from openapi_core.schema.operations.models import Operation
from openapi_core.schema.parameters.models import Parameter
from openapi_core.schema.paths.models import Path
from openapi_core.schema.schemas.enums import SchemaFormat, SchemaType
from openapi_core.schema.schemas.types import NoValue
from openapi_core.schema.servers.models import Server
from openapi_core.schema.specs.models import Spec
from openapi_core.templating.datatypes import TemplateResult
from openapi_core.templating.paths.exceptions import (
    OperationNotFound,
    PathNotFound,
    ServerNotFound,
)
from openapi_core.templating.paths.finders import PathFinder as CorePathFinder
from openapi_core.unmarshalling.schemas.enums import UnmarshalContext
from openapi_core.unmarshalling.schemas.exceptions import InvalidSchemaValue
from openapi_core.unmarshalling.schemas.factories import (
    SchemaUnmarshallersFactory as CoreSchemaUnmarshallersFactory,
)
from openapi_core.unmarshalling.schemas.formatters import Formatter
from openapi_core.unmarshalling.schemas.unmarshallers import (
    ArrayUnmarshaller as CoreArrayUnmarshaller,
    ObjectUnmarshaller as CoreObjectUnmarshaller,
)
from openapi_core.validation.request.datatypes import (
    OpenAPIRequest,
    RequestParameters,
)
from openapi_core.validation.request.validators import (
    RequestValidator as CoreRequestValidator,
)
from openapi_core.validation.response.datatypes import OpenAPIResponse
from openapi_core.validation.response.validators import (
    ResponseValidator as CoreResponseValidator,
)
from openapi_core.validation.validators import (
    BaseValidator as CoreBaseValidator,
)
from openapi_schema_validator._format import oas30_format_checker

from .data import OpenAPIParameters, to_openapi_parameters
from .exceptions import CastError, ValidationError
from .security import validate_security
from .utils import get_base_url
from ..annotations import MappingStrAny


DATE_TIME_FORMATTER = Formatter.from_callables(
    partial(oas30_format_checker.check, format="date-time"),
    parse_datetime,
)
PathTuple = Tuple[Path, Operation, Server, TemplateResult, TemplateResult]


class ArrayUnmarshaller(CoreArrayUnmarshaller):
    """Custom array unmarshaller to support nullable arrays.

    To be removed after ``openapi-core`` fixed an issue of supporting nullable
    arrays: https://github.com/p1c2u/openapi-core/issues/251
    """

    def __call__(self, value: Any = NoValue) -> Optional[List[Any]]:
        if value is None and self.schema.nullable:
            return None
        return cast(List[Any], super().__call__(value))


class EmailFormatter(Formatter):
    """Formatter to support email strings.

    Use `email-validator <https://pypi.org/project/email-validator>`_ library
    to ensure that given string is a valid email.
    """

    def validate(self, value: str) -> bool:
        try:
            validate_email(value)
        except EmailNotValidError as err:
            raise FormatError(f"{value!r} is not an 'email'", cause=err)
        return True


class ObjectUnmarshaller(CoreObjectUnmarshaller):
    """Custom object unmarshaller to support nullable objects.

    To be removed after ``openapi-core`` fixed an issue of` supporting nullable
    objects: https://github.com/p1c2u/openapi-core/issues/232
    """

    def __call__(self, value: Any = NoValue) -> Optional[Dict[Any, Any]]:
        if value is None and self.schema.nullable:
            return None
        return cast(Dict[Any, Any], super().__call__(value))


class PathFinder(CorePathFinder):
    """Custom path finder to fix issue with finding paths with vars.

    Temporary fix for https://github.com/p1c2u/openapi-core/issues/226, to be
    removed from *rororo* after next ``openapi-core`` version released with
    proper fix for the issue.
    """

    def find(self, request: OpenAPIRequest) -> PathTuple:
        """Better finder for request path.

        Instead of returning first possible result from
        ``self._get_servers_iter(...)`` call, attempt to ensure that
        ``request.full_url_pattern`` ends with ``path.name``.
        """
        paths_iter_peek = peekable(
            self._get_paths_iter(request.full_url_pattern)
        )
        if not paths_iter_peek:
            raise PathNotFound(request.full_url_pattern)

        operations_iter_peek = peekable(
            self._get_operations_iter(request.method, paths_iter_peek)
        )
        if not operations_iter_peek:
            raise OperationNotFound(request.full_url_pattern, request.method)

        servers_iter: Iterator[PathTuple] = self._get_servers_iter(
            request.full_url_pattern, operations_iter_peek
        )
        for server in servers_iter:
            path = server[0]
            if request.full_url_pattern.endswith(path.name):
                return server

        raise ServerNotFound(request.full_url_pattern)


class SchemaUnmarshallersFactory(CoreSchemaUnmarshallersFactory):
    """
    Custom schema unmarshallers factory to deal with tz aware date time
    strings.

    Temporary fix to https://github.com/p1c2u/openapi-core/issues/235, to be
    removed from *rororo* after next ``openapi-core`` release.
    """

    COMPLEX_UNMARSHALLERS = {
        **CoreSchemaUnmarshallersFactory.COMPLEX_UNMARSHALLERS,
        SchemaType.ARRAY: ArrayUnmarshaller,
        SchemaType.OBJECT: ObjectUnmarshaller,
    }

    def get_formatter(
        self,
        default_formatters: Dict[str, Formatter],
        type_format: str = None,
    ) -> Formatter:
        if type_format == SchemaFormat.DATETIME.value:
            return DATE_TIME_FORMATTER
        return super().get_formatter(default_formatters, type_format)


CUSTOM_FORMATTERS = {"email": EmailFormatter()}


class BaseValidator(CoreBaseValidator):
    """Custom base validator to deal with tz aware date time strings.

    To be removed from *rororo* after next ``openapi-core`` version release.
    """

    def _cast(self, param_or_media_type: Any, value: Any) -> Any:
        try:
            return super()._cast(param_or_media_type, value)
        except CoreCastError as err:
            # Pass param or media type name to cast error
            raise CastError(
                name=param_or_media_type.name, value=err.value, type=err.type
            )

    def _find_path(self, request: OpenAPIRequest) -> PathTuple:
        return PathFinder(self.spec, base_url=self.base_url).find(request)

    def _unmarshal(
        self, param_or_media_type: Any, value: Any, context: UnmarshalContext
    ) -> Any:
        """Use custom unmarshallers factory for unmarsahlling data."""
        if not param_or_media_type.schema:
            return value

        unmarshallers_factory = SchemaUnmarshallersFactory(
            self.spec._resolver,
            self.custom_formatters,
            context=context,
        )
        unmarshaller = unmarshallers_factory.create(param_or_media_type.schema)

        try:
            return unmarshaller(value)
        except InvalidSchemaValue as err:
            # Modify invalid schema validation errors to include parameter name
            if isinstance(param_or_media_type, Parameter):
                param_name = param_or_media_type.name

                for schema_error in err.schema_errors:
                    schema_error.path = schema_error.relative_path = deque(
                        [param_name]
                    )

            raise err


class RequestValidator(BaseValidator, CoreRequestValidator):
    def _get_parameters(
        self, request: OpenAPIRequest, params: MappingStrAny
    ) -> Tuple[RequestParameters, List[CoreOpenAPIError]]:
        """
        Distinct parameters errors from body errors to supply proper validation
        error response.
        """
        parameters, errors = super()._get_parameters(request, params)
        if errors:
            raise ValidationError.from_request_errors(
                errors, base_loc=["parameters"]
            )
        return parameters, errors

    def _get_security(
        self, request: OpenAPIRequest, operation: Operation
    ) -> MappingStrAny:
        """
        Custom logic for validating request security, which support handling
        JWT tokens without decoding them from ``base64`` (as needed for
        basic authorization).

        Consider to remove from *rororo* after next ``openapi-core`` release.
        """
        return validate_security(self, request, operation)

    def _unmarshal(  # type: ignore
        self, param_or_media_type: Any, value: Any
    ) -> Any:
        return super()._unmarshal(
            param_or_media_type, value, UnmarshalContext.REQUEST
        )


class ResponseValidator(BaseValidator, CoreResponseValidator):
    def _unmarshal(  # type: ignore
        self, param_or_media_type: Any, value: Any
    ) -> Any:
        return super()._unmarshal(
            param_or_media_type, value, UnmarshalContext.RESPONSE
        )


def validate_core_request(
    spec: Spec, core_request: OpenAPIRequest
) -> Tuple[MappingStrAny, OpenAPIParameters, Any]:
    """
    Instead of validating request parameters & body in two calls, validate them
    at once with passing custom formatters.
    """
    validator = RequestValidator(
        spec,
        custom_formatters=CUSTOM_FORMATTERS,
        base_url=get_base_url(core_request),
    )
    result = validator.validate(core_request)

    if result.errors:
        raise ValidationError.from_request_errors(result.errors)

    return (
        result.security,
        to_openapi_parameters(result.parameters),
        pyrsistent.freeze(result.body),
    )


def validate_core_response(
    spec: Spec,
    core_request: OpenAPIRequest,
    core_response: OpenAPIResponse,
) -> Any:
    """Pass custom formatters for validating response data."""
    validator = ResponseValidator(
        spec,
        custom_formatters=CUSTOM_FORMATTERS,
        base_url=get_base_url(core_request),
    )
    result = validator.validate(core_request, core_response)

    if result.errors:
        raise ValidationError.from_response_errors(result.errors)

    return result.data



import logging
import re
from typing import Any, Dict, List, Union

import attr
from jsonschema.exceptions import ValidationError as JsonSchemaValidationError
from openapi_core.casting.schemas.exceptions import CastError as CoreCastError
from openapi_core.deserializing.exceptions import DeserializeError
from openapi_core.deserializing.parameters.exceptions import (
    EmptyParameterValue,
)
from openapi_core.exceptions import OpenAPIError as CoreOpenAPIError
from openapi_core.schema.media_types.exceptions import (
    InvalidContentType,
    OpenAPIMediaTypeError,
)
from openapi_core.schema.parameters.exceptions import (
    MissingParameter,
    MissingRequiredParameter,
    OpenAPIParameterError,
)
from openapi_core.schema.responses.exceptions import InvalidResponse
from openapi_core.unmarshalling.schemas.exceptions import (
    InvalidSchemaValue,
    UnmarshalError,
)
from openapi_core.validation.exceptions import InvalidSecurity

from ..annotations import DictStrAny, MappingStrStr, TypedDict


ERROR_FIELD_REQUIRED = "Field required"
ERROR_NOT_FOUND_TEMPLATE = "{label} not found"
ERROR_PARAMETER_EMPTY = "Empty parameter value"
ERROR_PARAMETER_INVALID = "Invalid parameter value"
ERROR_PARAMETER_MISSING = "Missing parameter"
ERROR_PARAMETER_REQUIRED = "Parameter required"
OBJECT_LABEL = "Object"

PathItem = Union[int, str]
DictPathItemAny = Dict[PathItem, Any]

logger = logging.getLogger(__name__)
required_message_re = re.compile(
    r"^'(?P<field_name>.*)' is a required property$"
)


@attr.s(hash=True)
class CastError(CoreCastError):
    name: str = attr.ib()
    value: str = attr.ib()
    type: str = attr.ib()  # noqa: A003, VNE003


class OpenAPIError(Exception):
    """Base exception for other OpenAPI exceptions."""

    default_message = "Unhandled OpenAPI error"
    default_headers: MappingStrStr = {}
    status = 500

    def __init__(
        self, message: str = None, *, headers: MappingStrStr = None
    ) -> None:
        super().__init__(message or self.default_message)
        self.headers = headers or self.default_headers


class ConfigurationError(OpenAPIError):
    """Wrong OpenAPI configuration."""

    default_message = "OpenAPI schema configuration error"


class ContextError(OpenAPIError):
    """Attempt to use context when it is not available."""

    default_message = "OpenAPI context missed in request"


class OperationError(OpenAPIError):
    """Invalid OpenAPI operation."""

    default_message = "OpenAPI operation not found"


class BadRequest(OpenAPIError):
    """Bad request basic error."""

    default_message = "Bad request"
    status = 400


class SecurityError(OpenAPIError):
    """Request is not secured, but should."""

    default_message = "Not authenticated"
    status = 403


class BasicSecurityError(SecurityError):
    """Basic authentication is not provided."""

    default_headers = {"www-authenticate": "basic"}
    status = 401


class InvalidCredentials(SecurityError):
    """Invalid credentials provided for authentication."""

    default_message = "Invalid credentials"
    status = 403


class BasicInvalidCredentials(InvalidCredentials):
    """Invalid credentials provided for basic authentication."""

    default_headers = {"www-authenticate": "basic"}
    status = 401


class ObjectDoesNotExist(OpenAPIError):
    """Object does not exist basic error."""

    default_message = ERROR_NOT_FOUND_TEMPLATE.format(label=OBJECT_LABEL)
    status = 404

    def __init__(
        self,
        label: str = OBJECT_LABEL,
        *,
        message: str = None,
        headers: MappingStrStr = None,
    ) -> None:
        super().__init__(
            message or ERROR_NOT_FOUND_TEMPLATE.format(label=label),
            headers=headers,
        )
        self.label = label


class ValidationErrorItem(TypedDict):
    loc: List[PathItem]
    message: str


class ValidationError(OpenAPIError):
    """Request / response validation error.

    There is an ability to wrap ``openapi-core`` request / response validation
    errors via :func:`.from_request_errors` class method as in same time to
    create a validation error from the dict, like:

    .. code-block:: python

        raise ValidationError.from_dict(
            body={"name": "Name is not unique"}
        )
        raise ValidationError.from_dict(
            parameters={"X-Api-Key": "Outdated API key"}
        )
        raise ValidationError.from_dict(
            body={
                0: {
                    "data": {
                        0: {"name": "Field required"},
                        1: {"description": "Field required"},
                    },
                },
            },
        )

    Given interface is recommended for end users, who want to raise a
    custom validation errors within their operation handlers.
    """

    default_message = "Validation error"
    status = 422

    def __init__(
        self,
        *,
        message: str = None,
        errors: List[ValidationErrorItem] = None,
    ) -> None:
        super().__init__(message or self.default_message)

        self.errors = errors
        self.data = {"detail": errors} if errors else None

    @classmethod
    def from_dict(  # type: ignore
        cls, data: DictPathItemAny = None, **kwargs: Any
    ) -> "ValidationError":
        if data and kwargs:
            raise ValueError(
                "Please supply only data dict or kwargs, not both"
            )

        def dict_walker(
            loc: List[PathItem],
            data: Union[DictPathItemAny, DictStrAny],
            errors: List[ValidationErrorItem],
        ) -> None:
            for key, value in data.items():
                if isinstance(value, dict):
                    dict_walker(loc + [key], value, errors)
                else:
                    errors.append({"loc": loc + [key], "message": value})

        errors: List[ValidationErrorItem] = []
        dict_walker([], data or kwargs, errors)
        return cls(errors=errors)

    @classmethod
    def from_request_errors(  # type: ignore
        cls,
        errors: List[CoreOpenAPIError],
        *,
        base_loc: List[PathItem] = None,
    ) -> "ValidationError":
        base_loc = ["body"] if base_loc is None else base_loc
        result = []

        for err in errors:
            if isinstance(err, (OpenAPIParameterError, EmptyParameterValue)):
                result.append(get_parameter_error_details(base_loc, err))
            elif isinstance(err, OpenAPIMediaTypeError):
                result.append(get_media_type_error_details(base_loc, err))
            elif isinstance(err, CastError):
                result.append(
                    {
                        "loc": [*base_loc, err.name],
                        "message": (
                            f"{err.value!r} is not a type of {err.type!r}"
                        ),
                    }
                )
            elif isinstance(err, UnmarshalError):
                result.extend(get_unmarshal_error_details(base_loc, err))
            else:
                result.append(get_common_error_details(base_loc, err))

        return cls(
            message="Request parameters or body validation error",
            errors=result,
        )

    @classmethod
    def from_response_errors(  # type: ignore
        cls, errors: List[CoreOpenAPIError]
    ) -> "ValidationError":
        result: List[ValidationErrorItem] = []
        loc: List[PathItem] = ["response"]

        for err in errors:
            if isinstance(err, InvalidResponse):
                available_responses = ", ".join(sorted(err.responses))
                result.append(
                    {
                        "loc": loc,
                        "message": (
                            f"{err}. Available response http statuses: "
                            f"{available_responses}"
                        ),
                    }
                )
            elif isinstance(err, OpenAPIMediaTypeError):
                result.append(get_media_type_error_details(loc, err))
            elif isinstance(err, UnmarshalError):
                result.extend(get_unmarshal_error_details(loc, err))
            else:
                result.append(get_common_error_details(loc, err))

        return cls(message="Response data validation error", errors=result)


def ensure_loc(loc: List[PathItem]) -> List[PathItem]:
    return [item for item in loc if item != ""]


def get_common_error_details(
    loc: List[PathItem], err: CoreOpenAPIError
) -> ValidationErrorItem:
    return {"loc": loc, "message": str(err)}


def get_json_schema_validation_error_details(
    loc: List[PathItem], err: JsonSchemaValidationError
) -> ValidationErrorItem:
    message = err.message
    path = list(err.absolute_path)
    matched = required_message_re.match(message)

    if matched:
        return {
            "loc": ensure_loc(
                loc + path + [matched.groupdict()["field_name"]]
            ),
            "message": ERROR_FIELD_REQUIRED,
        }

    return {
        "loc": ensure_loc(loc + path),
        "message": err.message,
    }


def get_media_type_error_details(
    loc: List[PathItem], err: OpenAPIMediaTypeError
) -> ValidationErrorItem:
    if isinstance(err, InvalidContentType):
        return {
            "loc": loc,
            "message": (
                f"Schema missing for following mimetype: {err.mimetype}"
            ),
        }

    return get_common_error_details(loc, err)


def get_parameter_error_details(
    loc: List[PathItem],
    err: OpenAPIParameterError,
) -> ValidationErrorItem:
    parameter_name: str = getattr(err, "name", None)
    if parameter_name is None:
        return get_common_error_details(loc, err)

    message = {
        DeserializeError: ERROR_PARAMETER_INVALID,
        EmptyParameterValue: ERROR_PARAMETER_EMPTY,
        MissingParameter: ERROR_PARAMETER_MISSING,
        MissingRequiredParameter: ERROR_PARAMETER_REQUIRED,
    }[type(err)]

    return {"loc": [*loc, parameter_name], "message": message}


def get_unmarshal_error_details(
    loc: List[PathItem], err: UnmarshalError
) -> List[ValidationErrorItem]:
    if isinstance(err, InvalidSchemaValue):
        return [
            get_json_schema_validation_error_details(loc, item)
            for item in err.schema_errors
        ]

    return [get_common_error_details(loc, err)]


def is_only_security_error(errors: List[CoreOpenAPIError]) -> bool:
    return len(errors) == 1 and isinstance(errors[0], InvalidSecurity)



"""
==============
rororo.openapi
==============

Cornerstone of *rororo* library, which brings OpenAPI 3 schema support for
aiohttp.web applications.

"""

from .contexts import openapi_context
from .openapi import OperationTableDef, setup_openapi
from .utils import (
    get_openapi_context,
    get_openapi_schema,
    get_openapi_spec,
    get_validated_data,
    get_validated_parameters,
)
from .views import default_error_handler


__all__ = (
    "default_error_handler",
    "get_openapi_context",
    "get_openapi_schema",
    "get_openapi_spec",
    "get_validated_data",
    "get_validated_parameters",
    "openapi_context",
    "OperationTableDef",
    "setup_openapi",
)



"""
===============
rororo.settings
===============

Useful functions to work with application settings such as,

- Locale
- Logging
- Time zone

As well as provide attrib factory helper to read settings from environment to
use within Settings data structures.

"""

import calendar
import locale
import logging
import os
import time
import types
from importlib import import_module
from logging.config import dictConfig
from typing import (
    Any,
    Collection,
    Iterator,
    MutableMapping,
    Optional,
    Tuple,
    Type,
    Union,
)

import environ
from aiohttp import web

from .annotations import (
    DictStrAny,
    Level,
    MappingStrAny,
    MappingStrStr,
    Settings,
    T,
)
from .logger import default_logging_dict
from .utils import ensure_collection, to_bool


APP_SETTINGS_KEY = "settings"


@environ.config(prefix=None, frozen=True)
class BaseSettings:
    """Base Settings data structure for configuring ``aiohttp.web`` apps.

    Provides common attribs, which covers most of settings requires for
    run and configure ``aiohttp.web`` app. In same time it is designed to be
    inherited and completed with missed values in application as,

    .. code-block:: python

        import environ
        from rororo.settings import BaseSettings


        @environ.config(prefix=None, frozen=True)
        class Settings(BaseSettings):
            other_name: str = environ.var(
                name="OTHER_NAME", default="other-value"
            )

    """

    # Base aiohttp settings
    host: str = environ.var(name="AIOHTTP_HOST", default="localhost")
    port: int = environ.var(name="AIOHTTP_PORT", converter=int, default=8080)

    # Base application settings
    debug: bool = environ.bool_var(name="DEBUG", default=False)
    level: Level = environ.var(name="LEVEL", default="dev")

    # Date & time settings
    time_zone: str = environ.var(name="TIME_ZONE", default="UTC")

    # Locale settings
    first_weekday: int = environ.var(
        name="FIRST_WEEKDAY", converter=int, default=0
    )
    locale: str = environ.var(name="LOCALE", default="en_US.UTF-8")

    # Sentry settings
    sentry_dsn: Optional[str] = environ.var(name="SENTRY_DSN", default=None)
    sentry_release: Optional[str] = environ.var(
        name="SENTRY_RELEASE", default=None
    )

    def apply(
        self,
        *,
        loggers: Collection[str] = None,
        remove_root_handlers: bool = False,
    ) -> None:
        """
        Apply settings by calling setup logging, locale & timezone functions.

        Should be called once on application lifecycle. Best way to do it, to
        call right after settings instantiation.

        When ``loggers`` is passed, setup default logging dict for given
        iterable and call setup logging function. When ``loggers`` is omit do
        nothing.
        """
        if loggers:
            setup_logging(
                default_logging_dict(*ensure_collection(loggers)),
                remove_root_handlers=remove_root_handlers,
            )

        setup_locale(self.locale, self.first_weekday)
        setup_timezone(self.time_zone)

    @property
    def is_dev(self) -> bool:
        return self.level == "dev"

    @property
    def is_prod(self) -> bool:
        return self.level == "prod"

    @property
    def is_staging(self) -> bool:
        return self.level == "staging"

    @property
    def is_test(self) -> bool:
        return self.level == "test"


def from_env(key: str, default: T = None) -> Union[str, Optional[T]]:
    """Shortcut for safely reading environment variable.

    .. deprecated:: 2.0
        Use :func:`os.getenv` instead. Will be removed in **3.0**.

    :param key: Environment var key.
    :param default:
        Return default value if environment var not found by given key. By
        default: ``None``
    """
    return os.getenv(key, default)


def immutable_settings(defaults: Settings, **optionals: Any) -> MappingStrAny:
    r"""Initialize and return immutable Settings dictionary.

    Settings dictionary allows you to setup settings values from multiple
    sources and make sure that values cannot be changed, updated by anyone else
    after initialization. This helps keep things clear and not worry about
    hidden settings change somewhere around your web application.

    .. deprecated:: 2.0
        Function deprecated in favor or using `attrs <https://www.attrs.org>`_
        or `dataclasses <https://docs.python.org/3/library/dataclasses.html>`_
        for declaring settings classes. Will be removed in **3.0**.

    :param defaults:
       Read settings values from module or dict-like instance.
    :param \*\*optionals:
        Update base settings with optional values.

        In common additional values shouldn't be passed, if settings values
        already populated from local settings or environment. But in case
        of using application factories this makes sense::

            from . import settings

            def create_app(**options):
                app = ...
                app.settings = immutable_settings(settings, **options)
                return app

        And yes each additional key overwrite default setting value.
    """
    settings = {key: value for key, value in iter_settings(defaults)}
    for key, value in iter_settings(optionals):
        settings[key] = value
    return types.MappingProxyType(settings)


def inject_settings(
    mixed: Union[str, Settings],
    context: MutableMapping[str, Any],
    fail_silently: bool = False,
) -> None:
    """Inject settings values to given context.

    :param mixed:
        Settings can be a string (that it will be read from Python path),
        Python module or dict-like instance.
    :param context:
        Context to assign settings key values. It should support dict-like item
        assingment.
    :param fail_silently:
        When enabled and reading settings from Python path ignore errors if
        given Python path couldn't be loaded.
    """
    if isinstance(mixed, str):
        try:
            mixed = import_module(mixed)
        except Exception:
            if fail_silently:
                return
            raise

    for key, value in iter_settings(mixed):
        context[key] = value


def is_setting_key(key: str) -> bool:
    """Check whether given key is valid setting key or not.

    Only public uppercase constants are valid settings keys, all other keys
    are invalid and shouldn't present in Settings dict.

    **Valid settings keys**

    ::

        DEBUG
        SECRET_KEY

    **Invalid settings keys**

    ::

        _PRIVATE_SECRET_KEY
        camelCasedSetting
        rel
        secret_key

    :param key: Key to check.
    """
    return key.isupper() and key[0] != "_"


def iter_settings(mixed: Settings) -> Iterator[Tuple[str, Any]]:
    """Iterate over settings values from settings module or dict-like instance.

    :param mixed: Settings instance to iterate.
    """
    if isinstance(mixed, types.ModuleType):
        for item in dir(mixed):
            if not is_setting_key(item):
                continue
            yield (item, getattr(mixed, item))
    else:
        yield from filter(lambda item: is_setting_key(item[0]), mixed.items())


def setup_locale(
    lc_all: str,
    first_weekday: int = None,
    *,
    lc_collate: str = None,
    lc_ctype: str = None,
    lc_messages: str = None,
    lc_monetary: str = None,
    lc_numeric: str = None,
    lc_time: str = None,
) -> str:
    """Shortcut helper to setup locale for backend application.

    :param lc_all: Locale to use.
    :param first_weekday:
        Weekday for start week. 0 for Monday, 6 for Sunday. By default: None
    :param lc_collate: Collate locale to use. By default: ``<lc_all>``
    :param lc_ctype: Ctype locale to use. By default: ``<lc_all>``
    :param lc_messages: Messages locale to use. By default: ``<lc_all>``
    :param lc_monetary: Monetary locale to use. By default: ``<lc_all>``
    :param lc_numeric: Numeric locale to use. By default: ``<lc_all>``
    :param lc_time: Time locale to use. By default: ``<lc_all>``
    """
    if first_weekday is not None:
        calendar.setfirstweekday(first_weekday)

    locale.setlocale(locale.LC_COLLATE, lc_collate or lc_all)
    locale.setlocale(locale.LC_CTYPE, lc_ctype or lc_all)
    locale.setlocale(locale.LC_MESSAGES, lc_messages or lc_all)
    locale.setlocale(locale.LC_MONETARY, lc_monetary or lc_all)
    locale.setlocale(locale.LC_NUMERIC, lc_numeric or lc_all)
    locale.setlocale(locale.LC_TIME, lc_time or lc_all)

    return locale.setlocale(locale.LC_ALL, lc_all)


def setup_logging(
    config: DictStrAny, *, remove_root_handlers: bool = False
) -> None:
    """Wrapper around :func:`logging.config.dictConfig` function.

    In most cases it is not necessary to use an additional wrapper for setting
    up logging, but if your ``aiohttp.web`` application run as::

        python -m aiohttp.web api.app:create_app

    ``aiohttp`` `will setup
    <https://github.com/aio-libs/aiohttp/blob/v3.6.2/aiohttp/web.py#L494>`_
    logging via :func:`logging.basicConfig` call and it may result in
    duplicated logging messages. To avoid duplication, it is needed to remove
    ``logging.root`` handlers.

    :param remove_root_handlers:
        Remove ``logging.root`` handlers if any. By default: ``False``
    """
    if remove_root_handlers:
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)
    return dictConfig(config)


def setup_settings(
    app: web.Application,
    settings: BaseSettings,
    *,
    loggers: Collection[str] = None,
    remove_root_handlers: bool = False,
) -> web.Application:
    """Shortcut for applying settings for given ``aiohttp.web`` app.

    After applying, put settings to :class:`aiohttp.web.Application` dict as
    ``"settings"`` key.
    """
    settings.apply(loggers=loggers, remove_root_handlers=remove_root_handlers)
    app[APP_SETTINGS_KEY] = settings
    return app


def setup_settings_from_environ(
    app: web.Application,
    settings_class: Type[BaseSettings],
    *,
    environ: MappingStrStr = None,
    loggers: Collection[str] = None,
    remove_root_handlers: bool = False,
) -> web.Application:
    """
    Shortcut for instantiating settings from environ and applying them for
    given ``aiohttp.web`` app.

    This function calls ``settings_class.from_environ()`` method for you.

    After applying, put settings to :class:`aiohttp.web.Application` dict as
    ``"settings"`` key.
    """
    return setup_settings(
        app,
        settings_class.from_environ(environ or os.environ),  # type: ignore
        loggers=loggers,
        remove_root_handlers=remove_root_handlers,
    )


def setup_timezone(timezone: str) -> None:
    """Shortcut helper to configure timezone for backend application.

    :param timezone: Timezone to use, e.g. "UTC", "Europe/Kiev".
    """
    if timezone and hasattr(time, "tzset"):
        tz_root = "/usr/share/zoneinfo"
        tz_filename = os.path.join(tz_root, *(timezone.split("/")))

        if os.path.exists(tz_root) and not os.path.exists(tz_filename):
            raise ValueError("Incorrect timezone value: {0}".format(timezone))

        os.environ["TZ"] = timezone
        time.tzset()


# Make flake8 happy
(setup_logging, to_bool)



"""
=============
rororo.logger
=============

Logging utilities to simplify setting up Python logging.

"""

import logging
import sys
from typing import Any, Optional, Union

from .annotations import DictStrAny


class IgnoreErrorsFilter(object):
    """Ignore all warnings and errors from stdout handler."""

    def filter(self, record: logging.LogRecord) -> bool:  # noqa: A003
        """Allow only debug and info log messages to stdout handler."""
        return record.levelname in {"DEBUG", "INFO"}


def default_logging_dict(*loggers: str, **kwargs: Any) -> DictStrAny:
    r"""Prepare logging dict for :func:`logging.config.dictConfig`.

    *rororo* minds to simplify and unify logging configuration for
    ``aiohttp.web`` applications and cause of that the resulted logging config
    will:

    - Only messages from ``loggers`` will be processed
    - Pass all ``DEBUG`` & ``INFO`` logging messages to ``stdout``
    - Pass all other messages to ``stderr``
    - Any logging message will be formatted as:
      ``"%(asctime)s [%(levelname)s:%(name)s] %(message)s"``

    For example, to enable logging for ``aiohttp`` & ``api`` loggers,

    .. code-block:: python

        from logging.config import dictConfig

        dictConfig(default_logging_dict("aiohttp", "api"))

    :param \*loggers: Enable logging for each logger in sequence.
    :param \*\*kwargs: Setup additional logger params via keyword arguments.
    """
    kwargs.setdefault("level", "INFO")
    return {
        "version": 1,
        "disable_existing_loggers": True,
        "filters": {"ignore_errors": {"()": IgnoreErrorsFilter}},
        "formatters": {
            "default": {
                "format": "%(asctime)s [%(levelname)s:%(name)s] %(message)s"
            },
            "naked": {"format": u"%(message)s"},
        },
        "handlers": {
            "stdout": {
                "class": "logging.StreamHandler",
                "filters": ["ignore_errors"],
                "formatter": "default",
                "level": "DEBUG",
                "stream": sys.stdout,
            },
            "stderr": {
                "class": "logging.StreamHandler",
                "formatter": "default",
                "level": "WARNING",
                "stream": sys.stderr,
            },
        },
        "loggers": {
            logger: dict(handlers=["stdout", "stderr"], **kwargs)
            for logger in loggers
        },
    }


def update_sentry_logging(
    logging_dict: DictStrAny,
    sentry_dsn: Optional[str],
    *loggers: str,
    level: Union[str, int] = None,
    **kwargs: Any
) -> None:
    r"""Enable Sentry logging if Sentry DSN passed.

    .. deprecated:: 2.0
        Deprecated in favor of
        `sentry-sdk <https://pypi.org/project/sentry-sdk>`_ and will be removed
        in **3.0**.

    .. note::
        Sentry logging requires `raven <http://pypi.org/project/raven>`_
        library to be installed.

    **Usage**

    .. code-block:: python

        from logging.config import dictConfig

        LOGGING = default_logging_dict()
        SENTRY_DSN = "..."

        update_sentry_logging(LOGGING, SENTRY_DSN)
        dictConfig(LOGGING)

    **Using AioHttpTransport for SentryHandler**

    This will allow to use ``aiohttp.client`` for pushing data to Sentry in
    your ``aiohttp.web`` app, which means elimination of sync calls to Sentry.

    .. code-block:: python

        from raven_aiohttp import AioHttpTransport

        update_sentry_logging(
            LOGGING, SENTRY_DSN, transport=AioHttpTransport
        )

    :param logging_dict: Logging dict.
    :param sentry_dsn:
        Sentry DSN value. If ``None`` do not update logging dict at all.
    :param \*loggers:
        Use Sentry logging for each logger in the sequence. If the sequence is
        empty use Sentry logging to each available logger.
    :param \*\*kwargs: Additional kwargs to be passed to ``SentryHandler``.
    """
    # No Sentry DSN, nothing to do
    if not sentry_dsn:
        return

    # Add Sentry handler
    kwargs["class"] = "raven.handlers.logging.SentryHandler"
    kwargs["dsn"] = sentry_dsn
    logging_dict["handlers"]["sentry"] = dict(
        level=level or "WARNING", **kwargs
    )

    loggers = tuple(logging_dict["loggers"]) if not loggers else loggers
    for logger in loggers:
        # Ignore missing loggers
        logger_dict = logging_dict["loggers"].get(logger)
        if not logger_dict:
            continue

        # Ignore logger from logger config
        if logger_dict.pop("ignore_sentry", False):
            continue

        # Handlers list should exist
        handlers = list(logger_dict.setdefault("handlers", []))
        handlers.append("sentry")
        logger_dict["handlers"] = tuple(handlers)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['rororo', 'rororo.openapi']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.1,<6.0',
 'aiohttp-middlewares>=1.1.0,<2.0.0',
 'aiohttp>=3.5,<4.0',
 'attrs>=19.1,<21',
 'email-validator>=1.0.5,<2.0.0',
 'environ-config>=20.1.0,<21.0.0',
 'isodate>=0.6.0,<0.7.0',
 'openapi-core>=0.13.3,<0.14.0',
 'pyrsistent>=0.16.0,<0.17.0']

extras_require = \
{':python_version < "3.8"': ['typing-extensions>=3.7.4,<4.0.0']}

setup_kwargs = {
    'name': 'rororo',
    'version': '2.0.2',
    'description': 'aiohttp.web OpenAPI 3 schema first server applications.',
    'long_description': '======\nrororo\n======\n\n.. image:: https://github.com/playpauseandstop/rororo/workflows/ci/badge.svg\n    :target: https://github.com/playpauseandstop/rororo/actions?query=workflow%3A%22ci%22\n    :alt: CI Workflow\n\n.. image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n    :target: https://github.com/pre-commit/pre-commit\n    :alt: pre-commit\n\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/psf/black\n    :alt: black\n\n.. image:: https://img.shields.io/pypi/v/rororo.svg\n    :target: https://pypi.org/project/rororo/\n    :alt: Latest Version\n\n.. image:: https://img.shields.io/pypi/pyversions/rororo.svg\n    :target: https://pypi.org/project/rororo/\n    :alt: Python versions\n\n.. image:: https://img.shields.io/pypi/l/rororo.svg\n    :target: https://github.com/playpauseandstop/rororo/blob/master/LICENSE\n    :alt: BSD License\n\n.. image:: https://coveralls.io/repos/playpauseandstop/rororo/badge.svg?branch=master&service=github\n    :target: https://coveralls.io/github/playpauseandstop/rororo\n    :alt: Coverage\n\n.. image:: https://readthedocs.org/projects/rororo/badge/?version=latest\n    :target: https://rororo.readthedocs.io/\n    :alt: Documentation\n\nImplement `aiohttp.web`_ `OpenAPI 3`_ server applications with schema first\napproach.\n\nAs well as bunch other utilities to build effective server applications with\n`Python`_ 3 & `aiohttp.web`_.\n\n* Works on `Python`_ 3.6+\n* Works with `aiohttp.web`_ 3.6+\n* BSD licensed\n* Source, issues, and pull requests `on GitHub\n  <https://github.com/playpauseandstop/rororo>`_\n\n.. _`OpenAPI 3`: https://spec.openapis.org/oas/v3.0.3\n.. _`aiohttp.web`: https://aiohttp.readthedocs.io/en/stable/web.html\n.. _`Python`: https://www.python.org/\n\nQuick Start\n===========\n\n*rororo* relies on valid OpenAPI 3 schema file (both JSON or YAML formats\nsupported).\n\nExample below, illustrates on how to handle operation ``hello_world`` from\n`openapi.yaml </tests/openapi.yaml>`_ schema file.\n\n.. code-block:: python\n\n    from pathlib import Path\n    from typing import List\n\n    from aiohttp import web\n    from rororo import (\n        openapi_context,\n        OperationTableDef,\n        setup_openapi,\n    )\n\n\n    operations = OperationTableDef()\n\n\n    @operations.register\n    async def hello_world(request: web.Request) -> web.Response:\n        with openapi_context(request) as context:\n            name = context.parameters.query.get("name", "world")\n            return web.json_response({"message": f"Hello, {name}!"})\n\n\n    def create_app(argv: List[str] = None) -> web.Application:\n        return setup_openapi(\n            web.Application(),\n            Path(__file__).parent / "openapi.yaml",\n            operations,\n            server_url="/api",\n        )\n\nSchema First Approach\n---------------------\n\nUnlike other popular Python OpenAPI 3 solutions, such as\n`Django REST Framework`_, `FastAPI`_,  `flask-apispec`_, or `aiohttp-apispec`_\n*rororo* **requires** you to provide valid `OpenAPI 3`_ schema first. This\nmakes *rororo* similar to `connexion`_, `pyramid_openapi3`_ and other schema\nfirst libraries.\n\n.. _`Django REST Framework`: https://www.django-rest-framework.org\n.. _`FastAPI`: https://fastapi.tiangolo.com\n.. _`flask-apispec`: https://flask-apispec.readthedocs.io\n.. _`aiohttp-apispec`: https://aiohttp-apispec.readthedocs.io\n.. _`connexion`: https://connexion.readthedocs.io\n.. _`pyramid_openapi3`: https://github.com/Pylons/pyramid_openapi3\n\nClass Based Views\n-----------------\n\n*rororo* supports `class based views <https://docs.aiohttp.org/en/stable/web_quickstart.html#aiohttp-web-class-based-views>`_\nas well. `Todo-Backend </examples/todobackend>`_ example illustrates how to use\nclass based views for OpenAPI 3 servers.\n\nIn snippet below, *rororo* expects that OpenAPI 3 schema contains operation \\\nID ``UserView.get``,\n\n.. code-block:: python\n\n    @operations.register\n    class UserView(web.View):\n        async def get(self) -> web.Response:\n            ...\n\nMore Examples\n-------------\n\nCheck `examples </examples>`_ folder to see other examples on how to build\naiohttp.web OpenAPI 3 server applications.\n',
    'author': 'Igor Davydenko',
    'author_email': 'iam@igordavydenko.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://igordavydenko.com/projects/#rororo',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
